; -*- Mode:Lisp; Package:CCL; -*-
;;;
;;;   Copyright (C) 1994-2001 Digitool, Inc
;;;   This file is part of Opensourced MCL.
;;;
;;;   Opensourced MCL is free software; you can redistribute it and/or
;;;   modify it under the terms of the GNU Lesser General Public
;;;   License as published by the Free Software Foundation; either
;;;   version 2.1 of the License, or (at your option) any later version.
;;;
;;;   Opensourced MCL is distributed in the hope that it will be useful,
;;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;   Lesser General Public License for more details.
;;;
;;;   You should have received a copy of the GNU Lesser General Public
;;;   License along with this library; if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;;

; edit-callers.lisp



(in-package :ccl)

(eval-when (:compile-toplevel :execute)
  ;; argument must be a variable
  (defmacro lfun-vector-attributes (lfunv)
    (declare (ignore lfunv))
    0)

  ; arg is a variable
  (defmacro lfun-vector-bits (lfunv)
    `(%svref ,lfunv (1- (uvsize ,lfunv))))
   
  
  )

(defun global-function-p (random &optional name)
  (let* ((thing random)
         (name (or name (function-name thing))))
    (and name
         (or (not (or (symbolp name)(and (consp name)(eq (car name) 'setf)))) ; maybe its (setf baz)
             (let ((fn  (fboundp name)))
               (and fn
                    (progn
		; maybe this is enough for both cases?
                      (or (eq thing fn)
                          (and (symbolp name)(eq thing (macro-function name))))))))
         name)))

(defvar *function-parent-table* nil)
(defvar *function-parent-pool* (%cons-pool))

(defun copying-gc-p () ; if nz copying gc is on
  nil)

(defun lfun-closure-p (lfun)
  (logbitp $lfbits-trampoline-bit (lfun-bits lfun)))

; make a macro ?
(defun puthash-parent (im fun)
  (when (functionp im) ; was (or (functionp im)(eq imtype $sym.fapply))
    (if (global-function-p fun)
      (setf (gethash im *function-parent-table*) fun)
      (let ((ht (gethash im *function-parent-table*)))
        (if (not ht)
          (setf (gethash im *function-parent-table*) fun)
          (unless (eq ht fun)
            (if (consp ht)
              (when (not (memq fun ht))(nconc ht (list fun)))
              (if (not (global-function-p ht))
                (setf (gethash im *function-parent-table*) (list ht fun))))))))))       

; all nil excludes swapped functions, default is include (can be very slow) (.739)
(defun callers (function &optional (all t) 
                         &aux cfun callers gccount retry loadp)
                                        ;(declare (special cfun function callers))
  (declare (optimize (speed 3)(safety 0)))

  (let ((*function-parent-table* nil))
    (if (and (symbolp function) (fboundp function))
        (setq cfun (symbol-function function)))
    (if (and (consp function)(eq (car function) 'setf))
        (let ((nm (cadr function)))
          (setq function  (or (%setf-method nm)
                              (and (symbolp nm)
                                   (setq nm (setf-function-name nm))
                                   (fboundp nm)
                                   nm)
                              function))))  
    (when (copying-gc-p) (setq gccount (full-gccount)))
    (flet ((do-it (fun)
                                        ;(declare (special fun))
             (when (and gccount (neq gccount (full-gccount)))
               (throw 'losing :lost))
             (let ((lfunv (%lfun-vector fun loadp)))
               (when lfunv
                 (let ((bits (lfun-vector-bits lfunv))) ;(lfun-bits fun)))
                   (declare (fixnum bits))
                   (unless (or (and (logbitp $lfbits-cm-bit bits)(not (logbitp $lfbits-method-bit bits))) ; combined method
                               (and (logbitp $lfbits-trampoline-bit bits)(lfun-closure-p fun)
                                    (not (global-function-p fun)))) ; closure (interp or compiled)
                     (if (logbitp $lfbits-evaluated-bit bits)
                         (when (callers-interp fun function cfun)
                           (push fun callers))
                         (when  t
                               (let* ((nm (lfun-vector-name lfunv))
                                      (globalp (global-function-p fun nm)))
                                 (flet ((do-imm (im)
                                          (when (and (or (eq function im)
                                                         (and cfun (eq cfun im)))
                                                     (neq im nm))                             
                                            (push fun callers)) 
                                          (when (functionp im) ; was (or (functionp im)(eq imtype $sym.fapply))
                                            (if globalp
                                                (setf (gethash im *function-parent-table*) fun)
                                                (let ((ht (gethash im *function-parent-table*)))
                                                  (if (not ht)
                                                      (setf (gethash im *function-parent-table*) fun)
                                                      (unless (eq ht fun)
                                                        (if (consp ht)
                                                            (when (not (memq fun ht))(nconc ht (list fun)))
                                                            (if (not (global-function-p ht))
                                                                (setf (gethash im *function-parent-table*) 
                                                                      (list ht fun)))))))))))
                                   (declare (dynamic-extent #'do-imm))                                
                                   (%map-lfimms lfunv #'do-imm )))))))))))
      (declare (dynamic-extent #'do-it))
      (unwind-protect
           (progn
             (let* ((pool *function-parent-pool*)
                    (tbl (pool.data pool)))
               (setf (pool.data pool) nil
                     *function-parent-table*
                     (if tbl
                         (clrhash tbl)
                         (make-hash-table :size 700 :test 'eq :weak :value))))
             (loop
                 (cond ((eq :lost (catch 'losing      
                                    (%map-lfuns #'do-it)))
                        (when retry (error "Callers is losing"))
                        (setq callers nil)
                        (setq retry t))
                       (t (return))))
             (when all (setq loadp t)(map-swapped-lfuns #'do-it))
                                        ;(%map-static-lfuns #'do-it t)  ; there are only 4 of these - is it worth the trouble?
                                        ; Get rid of garbage methods - do we like this?
                                        ; what about generic flet?    
             (delete-if #'(lambda (thing)
                            (or (functionp thing)
                                (and (typep thing 'method)
                                     (let ((gf (fboundp (method-name thing))))
                                       (not (and (typep gf 'standard-generic-function)
                                                 (memq thing (%gf-methods gf))))))))
                        (delete-duplicates (mapcar 'top-level-caller callers))))
        (setf (pool.data *function-parent-pool*) *function-parent-table*
              *function-parent-table* nil)))))



(defun top-level-caller (function &optional the-list)
  (or (global-function-p function)
      (pascal-function-p function)
      (let ((name (function-name function)))
        (and name (function-encapsulation name) name))
      (let ((caller function) next)
        (loop
          (setq next (gethash caller *function-parent-table*))
          (if  next           
            (cond ((consp next)
                   (when (null the-list)(push function the-list))
                   (return
                    (dolist (c next)
                      (when (not (memq c the-list))
                        (let ((res (top-level-caller c the-list)))
                          (when (and res (not (functionp res)))
                            (return res)))))))
                  (t (let ((res (global-function-p next)))
                       (when res (return res)))
                     (when (null the-list)(push function the-list))
                     (when (memq next the-list) (return))
                     (push next the-list)
                     (setq caller next)))
            (return caller))))
      function))

; in 3.x the function in pascal-functions calls the actual function
(defun pascal-function-p (function)
  (if (find function %pascal-functions%
            :test #'eq
            :key #'(lambda (elt)
                     (if (consp elt)
                       (let ((one (cdr elt)))
                         (when (and (eq (function-name one)(function-name function))
                                    (block blob
                                      (%map-lfimms (%lfun-vector one) #'(lambda (imm)
                                                                          (when (eq imm function)
                                                                            (return-from blob function))))))
                           function))
                       (if elt (aref elt 2)))))
    (function-name function)))





                  

(defun map-swapped-lfuns (function)
  (declare (ignore function)))





(defun callers-interp (function target ctarget)
  (if (typep function 'interpreted-lexical-closure)
    (callers-interp (closure-function function) target ctarget)
    (let* ((lfunv (%lfun-vector function))
           (nimms (%count-immrefs lfunv))
           (body (cddr (closure-def-expanded-form (evalenv-closure-def (%nth-immediate lfunv 0)))))
           calls-target)
      ; crock!!!!! - what is the right way to do this?
      (when (not (consp body))      
        (dotimes (i nimms)
          (let ((it (%nth-immediate lfunv i nimms)))
            (when (consp it)(setq body it)(return)))))
      (labels ((calls-in-progn (body)
                 (dolist (expr body)
                   (calls-in-expr expr)))
               (calls-in-expr (expr)
                 (when (consp expr)
                   (let ((car (car expr)))
                     (if (consp car)
                       (ecase (car car)
                         ; 34 special forms
                         ((block progn if tagbody progv locally unwind-protect
                                 multiple-value-list multiple-value-prog1
                                 without-interrupts)
                          (calls-in-progn (cdr expr)))
                         (catch (calls-in-progn (cddr expr)))
                         (multiple-value-call (calls-in-progn (cdr expr)))
                         ((the return-from throw) (calls-in-expr (third expr)))                       
                         ((%with-specials eval-when) (calls-in-progn (cddr expr)))
                         ((%local-ref %special-ref quote %special-declare go
                                      %closure-ref %special-bind))
                         (%local-fref
                          (puthash-parent (second car) function)
                          (calls-in-progn (cdr expr)))
                         (%init&bind (calls-in-expr (third expr)))
                         ((let let* compiler-let)
                          (let ((args (second expr)))
                            (dolist (a args)
                              (when (consp a) (calls-in-expr (second a))))
                            (calls-in-progn (cddr expr))))
                         ((flet labels macrolet symbol-macrolet)
                          (calls-in-progn (cddr expr)))
                         (setq
                          (do ((l (cdr expr)(cddr l)))
                              ((null l))
                            (calls-in-expr (second l))))
                         (function
                          (let ((fn (second expr)))
                            (if (symbolp fn)
                              (when (eq fn target)(setq calls-target t))
                              (progn
                                (when (eq target ctarget)(setq calls-target t))
                                (puthash-parent (second expr) function))))))
                       (case car
                         (%local-fref                        
                          (puthash-parent (second expr) function))
                         (quote)
                         (t
                          (when (eq car target)
                            (setq calls-target t))
                          (calls-in-progn (cdr expr)))))))))
        (calls-in-progn body)
        calls-target))))

; Calls function f with args (imm) on each immediate in lfv.

(defun %map-lfimms (lfv f)
  (let* ((n (- (uvsize lfv) 2)))
    (declare (fixnum n))
    (dotimes (i n)
      (funcall f (%svref lfv (%i+ 1 i))))))
         
    


(provide :edit-callers)
